#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/**
 * This macro checks return value of the CUDA runtime call and exits
 * the application if the call failed.
 */
#define CUDA_CHECK_RETURN(value) {											\
	cudaError_t _m_cudaStat = value;										\
	if (_m_cudaStat != cudaSuccess) {										\
		fprintf(stderr, "Error %s at line %d in file %s\n",					\
				cudaGetErrorString(_m_cudaStat), __LINE__, __FILE__);		\
		exit(1);															\
	} }


#define BLOCK_SIZE 256

__global__ 
void matrixMulKernel(float* A, float* B, float* C, int width) {
	int i;
	int row = blockIdx.y*blockDim.y+threadIdx.y;
	int col = blockIdx.x*blockDim.x+threadIdx.x;

	if ((row<width) && (col<width)) {
		float tmp = 0;
		for (i = 0; i < width; ++i) 
			tmp += A[row*width+i]*B[i*width+col];
		C[row*width+col] = tmp;
	}
}

int main(int argc, char const *argv[]) {
	int i,j;
	float *hA=NULL, *hB=NULL, *hC=NULL;
	float *dA=NULL, *dB=NULL, *dC=NULL;

	int width = 2048;

	int nelements = width*width;

	hA = (float*)malloc(nelements*sizeof(float));
	hB = (float*)malloc(nelements*sizeof(float));
	hC = (float*)malloc(nelements*sizeof(float));

	//host parameter mxs initialization
	srand(time(NULL));
	for (i = 0; i < width; ++i) {
		for (j = 0; j < width; ++j) {
			hA[i*width+j] = rand();
			hB[i*width+j] = rand();
		}
	}

	clock_t allocStart = clock();
	CUDA_CHECK_RETURN(cudaMalloc((void**)&dA, nelements*sizeof(float)));
	CUDA_CHECK_RETURN(cudaMalloc((void**)&dB, nelements*sizeof(float)));
	CUDA_CHECK_RETURN(cudaMalloc((void**)&dC, nelements*sizeof(float)));
	clock_t allocEnd = clock();

	clock_t cpyStart = clock();
	CUDA_CHECK_RETURN(cudaMemcpy(dA, hA, nelements*sizeof(float), cudaMemcpyHostToDevice));
	CUDA_CHECK_RETURN(cudaMemcpy(dB, hB, nelements*sizeof(float), cudaMemcpyHostToDevice));
	clock_t cpyEnd = clock();

	dim3 gridSize(ceil(nelements/(double)BLOCK_SIZE), 1, 1);
	dim3 blockSize(BLOCK_SIZE, 1, 1);

	clock_t mulStart = clock();
	matrixMulKernel<<<gridSize, blockSize>>>(dA, dB, dC, width);
	clock_t mulEnd = clock();

	CUDA_CHECK_RETURN(cudaMemcpy(hC, dC, nelements*sizeof(float), cudaMemcpyDeviceToHost));

	double mulElapsed = (mulEnd-mulStart)/(double)CLOCKS_PER_SEC;
	double cpyElapsed = (cpyEnd-cpyStart)/(double)CLOCKS_PER_SEC;
	double allocElapsed = (allocEnd-allocStart)/(double)CLOCKS_PER_SEC;

	printf("Matrix dimensions: %dx%d\n", width, width);
	printf("Compute time: %f s\n", mulElapsed);
	printf("Allocation time: %f s\n", allocElapsed);
	printf("Memory copy time: %f s\n", cpyElapsed);

	cudaFree(dC);
	cudaFree(dB);
	cudaFree(dA);
	free(hC);
	free(hB);
	free(hA);

	return 0;
}
